@@ -3,6 +3,7 @@ from __future__ import annotations
 import argparse
 import builtins
 import logging
+import os
 import sys
 from functools import partial
 from pathlib import Path
@@ -146,11 +147,20 @@ def _ensure_installed(
     if not exists:
         env_type = env_type or env_type or manager.get_environment().get_type()
         eprint(f"initializing build environment (type: {env_type.name})")
+    elif upgrade:
+        eprint("upgrading build environment")
+    elif reinstall:
+        eprint("reinstalling build environment")
 
     current_type = manager.get_environment().get_type()
-    if env_type is not None and exists and env_type != current_type:
-        eprint(f"re-initializing build environment (type changed: {current_type.name} → {env_type.name})")
-        install = True
+    if env_type is not None:
+        type_changed = exists and env_type != current_type
+        if not install and type_changed:
+            eprint(f"re-initializing build environment (type changed: {current_type.name} → {env_type.name})")
+            install = True
+            manager.remove()
+        elif install and type_changed:
+            eprint(f"note: build environment type changed as well ({current_type.name} → {env_type.name})")
 
     if not install and exists:
         metadata = manager.get_metadata()
@@ -167,13 +177,16 @@ def _ensure_installed(
         if not lockfile or upgrade:
             logger.info("installing from requirements")
             source = requirements
+            transitive = True
         else:
             logger.info("installing from lock file")
             source = lockfile.to_pinned_requirement_spec()
             lockfile = None
+            transitive = False
 
-        manager.remove()
-        manager.install(source, env_type)
+        manager.install(source, env_type, transitive)
+    else:
+        eprint(f"reusing build environment (type: {current_type.name})")
 
 
 def main() -> NoReturn:
@@ -237,7 +250,7 @@ def main() -> NoReturn:
         manager,
         requirements,
         lockfile,
-        env_options.reinstall,
+        env_options.reinstall or (os.getenv("KRAKENW_REINSTALL") == "1"),
         env_options.upgrade,
         env_options.use,
     )
